@@ -1,25 +1,15 @@
 /* SPDX-License-Identifier: GPL-2.0-only */
 #ifndef LABWC_FONT_H
 #define LABWC_FONT_H
+#include <pango/pango-font.h>
 
 struct lab_data_buffer;
 
-enum font_slant {
-	FONT_SLANT_NORMAL = 0,
-	FONT_SLANT_ITALIC,
-	FONT_SLANT_OBLIQUE
-};
-
-enum font_weight {
-	FONT_WEIGHT_NORMAL = 0,
-	FONT_WEIGHT_BOLD
-};
-
 struct font {
 	char *name;
 	int size;
-	enum font_slant slant;
-	enum font_weight weight;
+	PangoStyle slant;
+	PangoWeight weight;
 };
 
 struct _PangoFontDescription *font_to_pango_desc(struct font *font);
