@@ -201,9 +201,26 @@ fill_section(const char *content, struct wl_list *list, uint32_t *found_buttons)
 	g_strfreev(identifiers);
 }
 
+static void
+clear_title_layout(void)
+{
+	struct title_button *button, *button_tmp;
+	wl_list_for_each_safe(button, button_tmp, &rc.title_buttons_left, link) {
+		wl_list_remove(&button->link);
+		zfree(button);
+	}
+	wl_list_for_each_safe(button, button_tmp, &rc.title_buttons_right, link) {
+		wl_list_remove(&button->link);
+		zfree(button);
+	}
+	rc.title_layout_loaded = false;
+}
+
 static void
 fill_title_layout(char *content)
 {
+	clear_title_layout();
+
 	struct wl_list *sections[] = {
 		&rc.title_buttons_left,
 		&rc.title_buttons_right,
@@ -341,6 +358,16 @@ fill_window_rule(char *nodename, char *content, struct parser_state *state)
 	}
 }
 
+static void
+clear_window_switcher_fields(void)
+{
+	struct window_switcher_field *field, *field_tmp;
+	wl_list_for_each_safe(field, field_tmp, &rc.window_switcher.fields, link) {
+		wl_list_remove(&field->link);
+		osd_field_free(field);
+	}
+}
+
 static void
 fill_window_switcher_field(char *nodename, char *content, struct parser_state *state)
 {
@@ -828,15 +855,38 @@ set_font_attr(struct font *font, const char *nodename, const char *content)
 		font->size = atoi(content);
 	} else if (!strcmp(nodename, "slant")) {
 		if (!strcasecmp(content, "italic")) {
-			font->slant = FONT_SLANT_ITALIC;
+			font->slant = PANGO_STYLE_ITALIC;
 		} else if (!strcasecmp(content, "oblique")) {
-			font->slant = FONT_SLANT_OBLIQUE;
+			font->slant = PANGO_STYLE_OBLIQUE;
 		} else {
-			font->slant = FONT_SLANT_NORMAL;
+			font->slant = PANGO_STYLE_NORMAL;
 		}
 	} else if (!strcmp(nodename, "weight")) {
-		font->weight = !strcasecmp(content, "bold") ?
-			FONT_WEIGHT_BOLD : FONT_WEIGHT_NORMAL;
+		if (!strcasecmp(content, "thin")) {
+			font->weight = PANGO_WEIGHT_THIN;
+		} else if (!strcasecmp(content, "ultralight")) {
+			font->weight = PANGO_WEIGHT_ULTRALIGHT;
+		} else if (!strcasecmp(content, "light")) {
+			font->weight = PANGO_WEIGHT_LIGHT;
+		} else if (!strcasecmp(content, "semilight")) {
+			font->weight = PANGO_WEIGHT_SEMILIGHT;
+		} else if (!strcasecmp(content, "book")) {
+			font->weight = PANGO_WEIGHT_BOOK;
+		} else if (!strcasecmp(content, "medium")) {
+			font->weight = PANGO_WEIGHT_MEDIUM;
+		} else if (!strcasecmp(content, "semibold")) {
+			font->weight = PANGO_WEIGHT_SEMIBOLD;
+		} else if (!strcasecmp(content, "bold")) {
+			font->weight = PANGO_WEIGHT_BOLD;
+		} else if (!strcasecmp(content, "ultrabold")) {
+			font->weight = PANGO_WEIGHT_ULTRABOLD;
+		} else if (!strcasecmp(content, "heavy")) {
+			font->weight = PANGO_WEIGHT_HEAVY;
+		} else if (!strcasecmp(content, "ultraheavy")) {
+			font->weight = PANGO_WEIGHT_ULTRAHEAVY;
+		} else {
+			font->weight = PANGO_WEIGHT_NORMAL;
+		}
 	}
 }
 
@@ -1025,7 +1075,24 @@ entry(xmlNode *node, char *nodename, char *content, struct parser_state *state)
 		return;
 	}
 
-	/* handle the rest */
+	if (!strcasecmp(nodename, "prefix.desktops")) {
+		xstrdup_replace(rc.workspace_config.prefix, content ? content : "");
+		return;
+	}
+
+	/*
+	 * Nodenames where we want to honour !content have to be parsed above
+	 * this point. An example of this is:
+	 *
+	 *     <desktops>
+	 *       <prefix></prefix>
+	 *     </desktops>
+	 *
+	 * In the case of the <prefix> element having content, the node will be
+	 * processed twice; first for the element itself (with no content) and
+	 * then the content itself. In this situation xstrdup_replace() is
+	 * called twice, but the end result is the right one.
+	 */
 	if (!content) {
 		return;
 	}
@@ -1209,8 +1276,6 @@ entry(xmlNode *node, char *nodename, char *content, struct parser_state *state)
 		rc.workspace_config.popuptime = atoi(content);
 	} else if (!strcasecmp(nodename, "number.desktops")) {
 		rc.workspace_config.min_nr_workspaces = MAX(1, atoi(content));
-	} else if (!strcasecmp(nodename, "prefix.desktops")) {
-		xstrdup_replace(rc.workspace_config.prefix, content);
 	} else if (!strcasecmp(nodename, "popupShow.resize")) {
 		if (!strcasecmp(content, "Always")) {
 			rc.resize_indicator = LAB_RESIZE_INDICATOR_ALWAYS;
@@ -1267,8 +1332,10 @@ entry(xmlNode *node, char *nodename, char *content, struct parser_state *state)
 		rc.mag_height = atoi(content);
 	} else if (!strcasecmp(nodename, "initScale.magnifier")) {
 		set_float(content, &rc.mag_scale);
+		rc.mag_scale = MAX(1.0, rc.mag_scale);
 	} else if (!strcasecmp(nodename, "increment.magnifier")) {
 		set_float(content, &rc.mag_increment);
+		rc.mag_increment = MAX(0, rc.mag_increment);
 	} else if (!strcasecmp(nodename, "useFilter.magnifier")) {
 		set_bool(content, &rc.mag_filter);
 	}
@@ -1347,6 +1414,7 @@ xml_tree_walk(xmlNode *node, struct parser_state *state)
 			continue;
 		}
 		if (!strcasecmp((char *)n->name, "fields")) {
+			clear_window_switcher_fields();
 			state->in_window_switcher_field = true;
 			traverse(n, state);
 			state->in_window_switcher_field = false;
@@ -1406,8 +1474,8 @@ static void
 init_font_defaults(struct font *font)
 {
 	font->size = 10;
-	font->slant = FONT_SLANT_NORMAL;
-	font->weight = FONT_WEIGHT_NORMAL;
+	font->slant = PANGO_STYLE_NORMAL;
+	font->weight = PANGO_WEIGHT_NORMAL;
 }
 
 static void
@@ -1486,7 +1554,7 @@ rcxml_init(void)
 	rc.unsnap_threshold = 20;
 	rc.unmaximize_threshold = 150;
 
-	rc.snap_edge_range = 1;
+	rc.snap_edge_range = 10;
 	rc.snap_overlay_enabled = true;
 	rc.snap_overlay_delay_inner = 500;
 	rc.snap_overlay_delay_outer = 500;
@@ -1656,22 +1724,25 @@ deduplicate_key_bindings(void)
 	}
 }
 
-static struct {
-	enum window_switcher_field_content content;
-	int width;
-} fields[] = {
-	{ LAB_FIELD_TYPE, 25 },
-	{ LAB_FIELD_TRIMMED_IDENTIFIER, 25 },
-	{ LAB_FIELD_TITLE, 50 },
-	{ LAB_FIELD_NONE, 0 },
-};
-
 static void
 load_default_window_switcher_fields(void)
 {
-	struct window_switcher_field *field;
+	static const struct {
+		enum window_switcher_field_content content;
+		int width;
+	} fields[] = {
+#if HAVE_LIBSFDO
+		{ LAB_FIELD_ICON, 5 },
+		{ LAB_FIELD_DESKTOP_ENTRY_NAME, 30 },
+		{ LAB_FIELD_TITLE, 65 },
+#else
+		{ LAB_FIELD_DESKTOP_ENTRY_NAME, 30 },
+		{ LAB_FIELD_TITLE, 70 },
+#endif
+	};
 
-	for (int i = 0; fields[i].content != LAB_FIELD_NONE; i++) {
+	struct window_switcher_field *field;
+	for (size_t i = 0; i < ARRAY_SIZE(fields); i++) {
 		field = znew(*field);
 		field->content = fields[i].content;
 		field->width = fields[i].width;
@@ -1760,13 +1831,20 @@ post_processing(void)
 		if (!rc.workspace_config.prefix) {
 			rc.workspace_config.prefix = xstrdup(_("Workspace"));
 		}
+
+		struct buf b = BUF_INIT;
 		struct workspace *workspace;
 		for (int i = nr_workspaces; i < rc.workspace_config.min_nr_workspaces; i++) {
 			workspace = znew(*workspace);
-			workspace->name = strdup_printf("%s %d",
-				rc.workspace_config.prefix, i + 1);
+			if (!string_null_or_empty(rc.workspace_config.prefix)) {
+				buf_add_fmt(&b, "%s ", rc.workspace_config.prefix);
+			}
+			buf_add_fmt(&b, "%d", i + 1);
+			workspace->name = xstrdup(b.data);
 			wl_list_append(&rc.workspace_config.workspaces, &workspace->link);
+			buf_clear(&b);
 		}
+		buf_reset(&b);
 	}
 	if (rc.workspace_config.popuptime == INT_MIN) {
 		rc.workspace_config.popuptime = 1000;
@@ -1775,10 +1853,6 @@ post_processing(void)
 		wlr_log(WLR_INFO, "load default window switcher fields");
 		load_default_window_switcher_fields();
 	}
-
-	if (rc.mag_scale <= 0.0) {
-		rc.mag_scale = 1.0;
-	}
 }
 
 static void
@@ -1955,15 +2029,7 @@ rcxml_finish(void)
 	zfree(rc.workspace_config.prefix);
 	zfree(rc.tablet.output_name);
 
-	struct title_button *p, *p_tmp;
-	wl_list_for_each_safe(p, p_tmp, &rc.title_buttons_left, link) {
-		wl_list_remove(&p->link);
-		zfree(p);
-	}
-	wl_list_for_each_safe(p, p_tmp, &rc.title_buttons_right, link) {
-		wl_list_remove(&p->link);
-		zfree(p);
-	}
+	clear_title_layout();
 
 	struct usable_area_override *area, *area_tmp;
 	wl_list_for_each_safe(area, area_tmp, &rc.usable_area_overrides, link) {
@@ -2010,11 +2076,7 @@ rcxml_finish(void)
 
 	regions_destroy(NULL, &rc.regions);
 
-	struct window_switcher_field *field, *field_tmp;
-	wl_list_for_each_safe(field, field_tmp, &rc.window_switcher.fields, link) {
-		wl_list_remove(&field->link);
-		osd_field_free(field);
-	}
+	clear_window_switcher_fields();
 
 	struct window_rule *rule, *rule_tmp;
 	wl_list_for_each_safe(rule, rule_tmp, &rc.window_rules, link) {
