@@ -599,13 +599,6 @@ cursor_process_motion(struct server *server, uint32_t time, double *sx, double *
 		dnd_icons_move(seat, seat->cursor->x, seat->cursor->y);
 	}
 
-	if ((ctx.view || ctx.surface) && rc.focus_follow_mouse
-			&& server->input_mode
-				!= LAB_INPUT_STATE_WINDOW_SWITCHER) {
-		desktop_focus_view_or_surface(seat, ctx.view, ctx.surface,
-			rc.raise_on_focus);
-	}
-
 	struct mousebind *mousebind;
 	wl_list_for_each(mousebind, &rc.mousebinds, link) {
 		if (mousebind->mouse_event == MOUSE_ACTION_DRAG
@@ -621,8 +614,27 @@ cursor_process_motion(struct server *server, uint32_t time, double *sx, double *
 		}
 	}
 
-	return cursor_update_common(server, &ctx, time,
+	struct wlr_surface *old_focused_surface =
+		seat->seat->pointer_state.focused_surface;
+
+	bool notify = cursor_update_common(server, &ctx, time,
 		/* cursor_has_moved */ true, sx, sy);
+
+	struct wlr_surface *new_focused_surface =
+		seat->seat->pointer_state.focused_surface;
+
+	if (rc.focus_follow_mouse && new_focused_surface
+			&& old_focused_surface != new_focused_surface) {
+		/*
+		 * If followMouse=yes, update the keyboard focus when the
+		 * cursor enters a surface
+		 */
+		desktop_focus_view_or_surface(seat,
+			view_from_wlr_surface(new_focused_surface),
+			new_focused_surface, rc.raise_on_focus);
+	}
+
+	return notify;
 }
 
 static uint32_t
@@ -641,15 +653,10 @@ _cursor_update_focus(struct server *server)
 	struct cursor_context ctx = get_cursor_context(server);
 
 	if ((ctx.view || ctx.surface) && rc.focus_follow_mouse
-			&& !rc.focus_follow_mouse_requires_movement
-			&& server->input_mode
-				!= LAB_INPUT_STATE_WINDOW_SWITCHER) {
+			&& !rc.focus_follow_mouse_requires_movement) {
 		/*
 		 * Always focus the surface below the cursor when
 		 * followMouse=yes and followMouseRequiresMovement=no.
-		 *
-		 * We should ignore them while window-switching though, because
-		 * calling desktop_focus_view() un-minimizes previewed window.
 		 */
 		desktop_focus_view_or_surface(&server->seat, ctx.view,
 			ctx.surface, rc.raise_on_focus);
@@ -1511,8 +1518,6 @@ cursor_init(struct seat *seat)
 
 void cursor_finish(struct seat *seat)
 {
-	/* TODO: either clean up all the listeners or none of them */
-
 	wl_list_remove(&seat->cursor_motion.link);
 	wl_list_remove(&seat->cursor_motion_absolute.link);
 	wl_list_remove(&seat->cursor_button.link);
@@ -1527,6 +1532,7 @@ void cursor_finish(struct seat *seat)
 	wl_list_remove(&seat->request_cursor.link);
 	wl_list_remove(&seat->request_set_shape.link);
 	wl_list_remove(&seat->request_set_selection.link);
+	wl_list_remove(&seat->request_set_primary_selection.link);
 
 	wlr_xcursor_manager_destroy(seat->xcursor_manager);
 	wlr_cursor_destroy(seat->cursor);
