@@ -486,6 +486,7 @@ new_output_notify(struct wl_listener *listener, void *data)
 	wl_signal_add(&wlr_output->events.request_state, &output->request_state);
 
 	wl_list_init(&output->regions);
+	wl_array_init(&output->osd_scene.items);
 
 	/*
 	 * Create layer-trees (background, bottom, top and overlay) and
@@ -572,6 +573,15 @@ output_init(struct server *server)
 	output_manager_init(server);
 }
 
+static void output_manager_finish(struct server *server);
+
+void
+output_finish(struct server *server)
+{
+	wl_list_remove(&server->new_output.link);
+	output_manager_finish(server);
+}
+
 static void
 output_update_for_layout_change(struct server *server)
 {
@@ -890,6 +900,15 @@ output_manager_init(struct server *server)
 		&server->gamma_control_set_gamma);
 }
 
+static void
+output_manager_finish(struct server *server)
+{
+	wl_list_remove(&server->output_layout_change.link);
+	wl_list_remove(&server->output_manager_apply.link);
+	wl_list_remove(&server->output_manager_test.link);
+	wl_list_remove(&server->gamma_control_set_gamma.link);
+}
+
 struct output *
 output_from_wlr_output(struct server *server, struct wlr_output *wlr_output)
 {
