@@ -114,12 +114,17 @@ _osd_update(struct server *server)
 			wl_list_for_each(workspace, &server->workspaces.all, link) {
 				bool active =  workspace == server->workspaces.current;
 				set_cairo_color(cairo, server->theme->osd_label_text_color);
-				cairo_rectangle(cairo, x, margin,
-					rect_width - padding, rect_height);
-				cairo_stroke(cairo);
+				struct wlr_fbox fbox = {
+					.x = x,
+					.y = margin,
+					.width = rect_width,
+					.height = rect_height,
+				};
+				draw_cairo_border(cairo, fbox,
+					theme->osd_workspace_switcher_boxes_border_width);
 				if (active) {
 					cairo_rectangle(cairo, x, margin,
-						rect_width - padding, rect_height);
+						rect_width, rect_height);
 					cairo_fill(cairo);
 				}
 				x += rect_width + padding;
@@ -469,6 +474,8 @@ destroy_workspace(struct workspace *workspace)
 	wlr_scene_node_destroy(&workspace->tree->node);
 	zfree(workspace->name);
 	wl_list_remove(&workspace->link);
+	wl_list_remove(&workspace->on_cosmic.activate.link);
+	wl_list_remove(&workspace->on_ext.activate.link);
 
 	lab_cosmic_workspace_destroy(workspace->cosmic_workspace);
 	lab_ext_workspace_destroy(workspace->ext_workspace);
